import os
import sys
import re
import copy
import pyfastaq
from ariba import sequence_metadata, cdhit


class Error (Exception): pass

rename_sub_regex = re.compile(r'[^\w.-]')


class ReferenceData:
    def __init__(self,
        presence_absence_fa=None,
        variants_only_fa=None,
        non_coding_fa=None,
        metadata_tsv=None,
        min_gene_length=6,
        max_gene_length=10000,
        genetic_code=11,
    ):
        self.seq_filenames = {}
        self.seq_dicts = {}
        self.min_gene_length = min_gene_length
        self.max_gene_length = max_gene_length

        total_ref_seqs_loaded = 0

        for x in ['presence_absence', 'variants_only', 'non_coding']:
            exec('self.seq_filenames[x] = self._get_filename(' + x + '_fa)')
            self.seq_dicts[x] = self._load_fasta_file(self.seq_filenames[x])
            total_ref_seqs_loaded += len(self.seq_dicts[x])

        if {None} == set(self.seq_filenames.values()):
            raise Error('Error! Must supply at least one of presence_absence_fa, variants_only_fa, non_coding_fa. Cannot continue')

        if total_ref_seqs_loaded == 0:
            raise Error('Error! No sequences found in input file(s). Maybe they were empty? Cannot continue.')

        self.metadata = self._load_metadata_tsv(metadata_tsv)
        self.genetic_code = genetic_code
        pyfastaq.sequences.genetic_code = self.genetic_code
        common_names = self._dict_keys_intersection(list(self.seq_dicts.values()))
        if len(common_names):
            raise Error('Error! Non-unique names found in input fasta files:\n' + '\n'.join(common_names))


    @staticmethod
    def _dict_keys_intersection(dicts):
        dicts = [x for x in dicts if x is not None]
        if len(dicts) == 0:
            return set()

        inter = set(dicts[0].keys())

        for d in dicts[1:]:
            inter = inter.intersection(set(d.keys()))
        return inter


    @staticmethod
    def _get_filename(filename):
        if filename is None:
            return None
        else:
            if os.path.exists(filename):
                return os.path.abspath(filename)
            else:
                raise Error('Error! File not found: ' + filename)


    @staticmethod
    def _load_metadata_tsv(filename):
        if filename is None:
            return {}

        f = pyfastaq.utils.open_file_read(filename)
        metadata_dict = {}

        for line in f:
            try:
                metadata = sequence_metadata.SequenceMetadata(line)
            except:
                print('Problem with this line of metadata, which will be ignored:', line.rstrip(), file=sys.stderr)
                continue

            if metadata.name not in metadata_dict:
                metadata_dict[metadata.name] = {'n': {}, 'p': {}, '.': set()}

            if metadata.variant_type == '.':
                metadata_dict[metadata.name]['.'].add(metadata)
            else:
                if metadata.variant.position not in metadata_dict[metadata.name][metadata.variant_type]:
                    metadata_dict[metadata.name][metadata.variant_type][metadata.variant.position] = set()

                metadata_dict[metadata.name][metadata.variant_type][metadata.variant.position].add(metadata)

        pyfastaq.utils.close(f)
        return metadata_dict


    @staticmethod
    def _load_fasta_file(filename):
        d = {}

        if filename is not None:
            seq_reader = pyfastaq.sequences.file_reader(filename)
            for seq in seq_reader:
                #seq.id = seq.id.split()[0]
                if seq.id in d:
                    raise Error('Duplicate name "' + seq.id + '" found in file ' + filename + '. Cannot continue)')
                d[seq.id] = copy.copy(seq)

        return d


    @staticmethod
    def _find_gene_in_seqs(name, dicts):
        for dict_name, this_dict in dicts.items():
            if this_dict is None:
                continue
            elif name in this_dict:
                return dict_name

        return None


    @staticmethod
    def _write_metadata_tsv(metadata, filename):
        f = pyfastaq.utils.open_file_write(filename)

        for gene_name, data_dict in sorted(metadata.items()):
            for meta in sorted([str(x) for x in data_dict['.']]):
                print(meta, file=f)

            variants = []

            for variant_type in ['n', 'p']:
                for position in data_dict[variant_type]:
                    for meta in data_dict[variant_type][position]:
                        variants.append(meta)

            variants.sort()
            for v in variants:
                print(v, file=f)

        pyfastaq.utils.close(f)


    @staticmethod
    def _write_dict_of_sequences(seq_dict, filename):
        f = pyfastaq.utils.open_file_write(filename)
        for seq in sorted(seq_dict):
            print(seq_dict[seq], file=f)
        pyfastaq.utils.close(f)


    def _write_sequences(self, filename, sequences_to_write):
        assert sequences_to_write in self.seq_dicts and sequences_to_write in self.seq_filenames
        if self.seq_filenames[sequences_to_write] is not None:
            self._write_dict_of_sequences(self.seq_dicts[sequences_to_write], filename)


    def _filter_bad_variant_data(self, out_prefix, presence_absence_removed, variants_only_removed):
        genes_to_remove = set()
        variants_only_genes_not_found = set(self.seq_dicts['variants_only'].keys())
        log_file = out_prefix + '.log'
        tsv_file = out_prefix + '.tsv'
        log_fh = pyfastaq.utils.open_file_write(log_file)

        for gene_name, metadata_dict in sorted(self.metadata.items()):
            if gene_name in presence_absence_removed:
                print(gene_name, 'was removed from presence/absence fasta, so removing its metadata', file=log_fh)
                genes_to_remove.add(gene_name)
                continue
            elif gene_name in variants_only_removed:
                print(gene_name, 'was removed from variants only fasta, so removing its metadata', file=log_fh)
                genes_to_remove.add(gene_name)
                continue

            gene_in_seq_dict = self._find_gene_in_seqs(gene_name, self.seq_dicts)
            if gene_in_seq_dict is None:
                print(gene_name, 'is in input tsv file, but not found in any input sequence files. Removing', file=log_fh)
                genes_to_remove.add(gene_name)
                continue

            # take out any metadata that is not a variant and has no extra info.
            to_remove = []

            for metadata in metadata_dict['.']:
                if metadata.free_text == '.':
                    print(gene_name, 'metadata has no info. Just gene name given. Removing. Line of file was:', metadata, file=log_fh)
                    to_remove.append(metadata)

            for metadata in to_remove:
                metadata_dict['.'].remove(metadata)


            # if this is non_coding, we shouldn't have any amino acid variants
            if gene_in_seq_dict == 'non_coding':
                for position in metadata_dict['p']:
                    for metadata in metadata_dict['p'][position]:
                        print(gene_name, 'variant of type "p" for protein, but sequence is non-coding. Removing. Line of file was:', metadata, file=log_fh)

                metadata_dict['p'] = {}


            # take out variant metadata that doesn't make sense (eg bases not matching ref sequence)
            for variant_type in ['n', 'p']:
                positions_to_remove = []
                for position in metadata_dict[variant_type]:
                    meta_to_remove = []
                    for metadata in metadata_dict[variant_type][position]:
                        to_translate = variant_type == 'p'

                        if not metadata.variant.sanity_check_against_seq(self.seq_dicts[gene_in_seq_dict][gene_name], translate_seq=to_translate):
                            print(gene_name, 'variant does not match reference. Removing. Line of file was:', metadata, file=log_fh)
                            meta_to_remove.append(metadata)
                            continue

                        if gene_in_seq_dict == 'variants_only':
                            variants_only_genes_not_found.discard(gene_name)

                    for metadata in meta_to_remove:
                        metadata_dict[variant_type][position].remove(metadata)
                    if len(metadata_dict[variant_type][position]) == 0:
                        positions_to_remove.append(position)

                for position in positions_to_remove:
                    del metadata_dict[variant_type][position]


            if gene_in_seq_dict == 'variants_only' and len(metadata_dict['n']) == len(metadata_dict['p']) == len(metadata_dict['.']) == 0:
                print(gene_name, 'No remaining data after checks. Removing this sequence because it is in the variants only file', file=log_fh)
                genes_to_remove.add(gene_name)

        for gene_name in genes_to_remove:
            self.metadata.pop(gene_name)

        for gene_name in variants_only_genes_not_found:
            print(gene_name, 'is in variants only gene file, but no variants found. Removing.', file=log_fh)
            self.seq_dicts['variants_only'].pop(gene_name)

        pyfastaq.utils.close(log_fh)
        self._write_metadata_tsv(self.metadata, tsv_file)
        self._write_sequences(out_prefix + '.presence_absence.fa', 'presence_absence')
        self._write_sequences(out_prefix + '.non_coding.fa', 'non_coding')
        self._write_sequences(out_prefix + '.variants_only.fa', 'variants_only')


    @staticmethod
    def _try_to_get_gene_seq(seq, min_length, max_length):
        seq.seq = seq.seq.upper()
        if len(seq) < min_length:
            return None, 'Remove: too short. Length: ' + str(len(seq))
        elif len(seq) > max_length:
            return None, 'Remove: too long. Length: ' + str(len(seq))
        else:
            got = seq.make_into_gene()
            if got is None:
                return None, 'Does not look like a gene (tried both strands and all reading frames) ' + seq.seq
            else:
                return got[0], 'Made ' + seq.id + ' into gene. strand=' + got[1] + ', frame=' + str(got[2])


    def _remove_bad_genes(self, seqs_dict, log_file):
        to_remove = set()

        if len(seqs_dict) == 0:
            return to_remove

        log_fh = pyfastaq.utils.open_file_write(log_file)

        for name in sorted(seqs_dict):
            new_seq, message = self._try_to_get_gene_seq(seqs_dict[name], self.min_gene_length, self.max_gene_length)
            if new_seq is None:
                to_remove.add(name)
            else:
                seqs_dict[name] = new_seq

            if message is not None:
                print(name, message, file=log_fh)

        pyfastaq.utils.close(log_fh)

        for name in to_remove:
            seqs_dict.pop(name)

        return to_remove


    def sanity_check(self, outprefix):
        variants_only_removed = self._remove_bad_genes(self.seq_dicts['variants_only'], outprefix + '.00.check_fasta_variants_only.log')
        presence_absence_removed = self._remove_bad_genes(self.seq_dicts['presence_absence'], outprefix + '.00.check_fasta_presence_absence.log')
        self._filter_bad_variant_data(outprefix + '.01.check_variants', presence_absence_removed, variants_only_removed)


    @classmethod
    def _new_seq_name(cls, name):
        name = name.split()[0]
        return re.sub(rename_sub_regex, '_', name)


    @classmethod
    def _seq_names_to_rename_dict(cls, names):
        used_names = set()
        old_name_to_new = {}

        for old_name in sorted(names):
            new_name = ReferenceData._new_seq_name(old_name)
            if new_name in used_names:
                i = 1
                new_name_prefix = new_name
                while new_name in used_names:
                    new_name = new_name_prefix + '_' + str(i)
                    i += 1

            assert new_name not in used_names
            if new_name != old_name:
                old_name_to_new[old_name] = new_name

            used_names.add(new_name)

        return old_name_to_new


    @classmethod
    def _rename_names_in_seq_dicts(cls, seq_dicts, rename_dict):
        '''Changes seq_dicts in place'''
        for seq_type in ['presence_absence', 'variants_only', 'non_coding']:
            new_dict = {}
            while len(seq_dicts[seq_type]):
                old_name, seq = seq_dicts[seq_type].popitem()
                if old_name in rename_dict:
                    seq.id = rename_dict[old_name]

                new_dict[seq.id] = seq
            seq_dicts[seq_type] = new_dict


    @classmethod
    def _rename_metadata_set(cls, metadata_set, new_name):
        new_set = set()
        for meta in metadata_set:
            new_meta = copy.copy(meta)
            new_meta.name = new_name
            new_set.add(new_meta)
        return new_set


    @classmethod
    def _rename_names_in_metadata(cls, meta_dict, rename_dict):
        new_dict = {}

        while len(meta_dict):
            old_name, gene_dict = meta_dict.popitem()
            if old_name in rename_dict:
                new_name = rename_dict[old_name]
                for seq_type in ['n', 'p']:
                    for position, metaset in gene_dict[seq_type].items():
                        gene_dict[seq_type][position] = ReferenceData._rename_metadata_set(metaset, new_name)

                gene_dict['.'] = ReferenceData._rename_metadata_set(gene_dict['.'], new_name)
            else:
                new_name = old_name

            new_dict[new_name] = gene_dict

        return new_dict


    def rename_sequences(self, outfile):
        presabs_names = set(self.seq_dicts['presence_absence'].keys())
        noncoding_names = set(self.seq_dicts['non_coding'].keys())
        varonly_names = set(self.seq_dicts['variants_only'].keys())
        # we should have already checked that all the names are unique, but let's do it again!
        all_names = presabs_names.union(noncoding_names).union(varonly_names)
        if len(all_names) != len(presabs_names) + len(noncoding_names) + len(varonly_names):
            raise Error('Got a non-unique name in input data. Cannot continue')

        rename_dict = ReferenceData._seq_names_to_rename_dict(all_names)
        if len(rename_dict):
            print('Had to rename some sequences. See', outfile, 'for old -> new names', file=sys.stderr)
            with open(outfile, 'w') as f:
                for old_name, new_name in sorted(rename_dict.items()):
                    print(old_name, new_name, sep='\t', file=f)

            ReferenceData._rename_names_in_seq_dicts(self.seq_dicts, rename_dict)
            self.metadata = ReferenceData._rename_names_in_metadata(self.metadata, rename_dict)


    def make_catted_fasta(self, outfile):
        f = pyfastaq.utils.open_file_write(outfile)

        for key in ['presence_absence', 'variants_only', 'non_coding']:
            filename = self.seq_filenames[key]
            if filename is not None:
                file_reader = pyfastaq.sequences.file_reader(filename)
                for seq in file_reader:
                    print(seq, file=f)

        pyfastaq.utils.close(f)


    def sequence_type(self, sequence_name):
        return self._find_gene_in_seqs(sequence_name, self.seq_dicts)


    def sequence(self, sequence_name):
        d = self._find_gene_in_seqs(sequence_name, self.seq_dicts)
        if d is None:
            return None
        else:
            return self.seq_dicts[d][sequence_name]


    def sequence_length(self, sequence_name):
        seq = self.sequence(sequence_name)
        assert seq is not None
        return len(seq)


    def all_non_wild_type_variants(self, ref_name):
        ref_seq = self.sequence(ref_name)
        variants = {'n': {}, 'p': {}}

        if ref_seq is None or ref_name not in self.metadata:
            return variants

        for variant_type in ['n', 'p']:
            for position, metadata_set in self.metadata[ref_name][variant_type].items():
                for metadata in metadata_set:
                    if position not in variants[variant_type]:
                        variants[variant_type][position] = set()

                    variants[variant_type][position].add(metadata)

        return variants


    @staticmethod
    def write_cluster_allocation_file(clusters, outfile):
        f_out = pyfastaq.utils.open_file_write(outfile)

        for seq_type in ['presence_absence', 'variants_only', 'non_coding']:
            if clusters[seq_type] is None:
                continue

            for seq_name in sorted(clusters[seq_type]):
                other_seqs = clusters[seq_type][seq_name].difference({seq_name})
                if len(other_seqs) > 0:
                    other_seq_string = '\t'.join(sorted(list(other_seqs)))
                    print(seq_name, other_seq_string, sep='\t', file=f_out)
                else:
                    print(seq_name, file=f_out)

        pyfastaq.utils.close(f_out)


    def cluster_with_cdhit(self, inprefix, outprefix, seq_identity_threshold=0.9, threads=1, length_diff_cutoff=0.9, nocluster=False, verbose=False, cd_hit_est='cd-hit-est', clusters_file=None):
        files_to_cat = []
        clusters = {}

        for seqs_type in ['presence_absence', 'variants_only', 'non_coding']:
            if len(self.seq_dicts[seqs_type]) > 0:
                outfile = outprefix + '.' + seqs_type + '.cdhit'
                files_to_cat.append(outfile)
                cdhit_runner = cdhit.Runner(
                  inprefix + '.' + seqs_type + '.fa',
                  outfile,
                  seq_identity_threshold=seq_identity_threshold,
                  threads=threads,
                  length_diff_cutoff=length_diff_cutoff,
                  verbose=verbose,
                  cd_hit_est=cd_hit_est,
                  rename_suffix = seqs_type[0]
                )

                if clusters_file is not None:
                    new_clusters = cdhit_runner.run_get_clusters_from_file(clusters_file)
                elif nocluster:
                    new_clusters = cdhit_runner.fake_run()
                else:
                    new_clusters = cdhit_runner.run()

                clusters[seqs_type] = new_clusters
            else:
                clusters[seqs_type] = None

        assert len(files_to_cat) > 0
        f_out = pyfastaq.utils.open_file_write(outprefix + '.cluster_representatives.fa')

        for filename in files_to_cat:
            for seq in pyfastaq.sequences.file_reader(filename):
                print(seq, file=f_out)

        pyfastaq.utils.close(f_out)
        self.write_cluster_allocation_file(clusters, outprefix + '.clusters.tsv')
        return clusters


    def write_seqs_to_fasta(self, outfile, names):
        f_out = pyfastaq.utils.open_file_write(outfile)

        for name in sorted(names):
            print(self.sequence(name), file=f_out)

        pyfastaq.utils.close(f_out)



import argparse
import sys
import ariba

def run():
    parser = argparse.ArgumentParser(
        description = 'Filters an ARIBA report tsv file',
        usage = 'ariba reportfilter [options] <infile> <outprefix>'
    )
    parser.add_argument('--exclude_flags', help='Comma-separated list of flags to exclude. [%(default)s]', default='assembly_fail,ref_seq_choose_fail')
    parser.add_argument('--min_pc_id', type=float, help='Minimum percent identity of nucmer match between contig and reference [%(default)s]', default=90.0, metavar='FLOAT')
    parser.add_argument('--min_ref_base_asm', type=int, help='Minimum number of reference bases matching assembly [%(default)s]', default=1, metavar='INT')
    parser.add_argument('--keep_syn', action='store_true', help='Keep synonymous variants (by default they are removed')
    parser.add_argument('--discard_without_known_var', action='store_true', help='Applies to variant only genes. Filter out where there is a known variant, but the assembly has the wild type. By default these rows are kept.')
    parser.add_argument('infile', help='Name of input tsv file')
    parser.add_argument('outprefix', help='Prefix of output files. outprefix.tsv and outprefix.xls will be made')
    options = parser.parse_args()

    flags_to_exclude = options.exclude_flags.split(',')
    allowed_flags = set(ariba.flag.flags_in_order)
    bad_flags = [x for x in flags_to_exclude if x not in allowed_flags]
    if len(bad_flags):
        print('Error in option --exclude_flags. The following were not recognised:', ','.join(bad_flags), file=sys.stderr)
        print('Must choose from:', ','.join(ariba.flag.flags_in_order), file=sys.stderr)
        sys.exit(1)

    rf = ariba.report_filter.ReportFilter(
        infile=options.infile,
        min_pc_ident=options.min_pc_id,
        min_ref_base_assembled=options.min_ref_base_asm,
        ignore_not_has_known_variant=options.discard_without_known_var,
        remove_synonymous_snps=not options.keep_syn,
    )
    rf.run(options.outprefix)




import argparse
import ariba


def use_preset(options):
    if options.preset is None:
        return options

    preset_to_vals = {
        'minimal': {
            'cluster_cols': 'has_res',
            'variant_cols': '',
            'col_filter': 'y',
            'row_filter': 'y',
            'var_groups': 'n',
            'known_vars': 'n',
            'novel_vars': 'n'
        },
        'cluster_small': {
            'cluster_cols': 'assembled,has_res,ref_seq,known_var',
            'variant_cols': '',
            'col_filter': 'y',
            'row_filter': 'y',
            'var_groups': 'n',
            'known_vars': 'n',
            'novel_vars': 'n'
        },
        'cluster_all': {
            'cluster_cols': 'assembled,has_res,ref_seq,pct_id,known_var,novel_var',
            'variant_cols': '',
            'col_filter': 'y',
            'row_filter': 'y',
            'var_groups': 'n',
            'known_vars': 'n',
            'novel_vars': 'n'
        },
        'cluster_var_groups': {
            'cluster_cols': 'assembled,has_res,ref_seq,pct_id,known_var,novel_var',
            'variant_cols': 'groups',
            'col_filter': 'y',
            'row_filter': 'y',
            'var_groups': 'y',
            'known_vars': 'n',
            'novel_vars': 'n'
        },
        'cluster_known_vars': {
            'cluster_cols': 'assembled,has_res,ref_seq,pct_id,known_var,novel_var',
            'variant_cols': 'groups,grouped,ungrouped',
            'col_filter': 'y',
            'row_filter': 'y',
            'var_groups': 'y',
            'known_vars': 'y',
            'novel_vars': 'n'
        },
        'all': {
            'cluster_cols': 'assembled,has_res,ref_seq,pct_id,known_var,novel_var',
            'variant_cols': 'groups,grouped,ungrouped,novel',
            'col_filter': 'y',
            'row_filter': 'y',
            'var_groups': 'y',
            'known_vars': 'y',
            'novel_vars': 'y'
        },
        'all_no_filter': {
            'cluster_cols': 'assembled,has_res,ref_seq,pct_id,known_var,novel_var',
            'variant_cols': 'groups,grouped,ungrouped,novel',
            'col_filter': 'n',
            'row_filter': 'n',
            'var_groups': 'y',
            'known_vars': 'y',
            'novel_vars': 'y'
        },
    }

    assert options.preset in preset_to_vals

    for key, val in preset_to_vals[options.preset].items():
        exec('options.' + key + ' = "' + val + '"')

    return options


def run():
    presets = ['minimal', 'cluster_small', 'cluster_all', 'cluster_var_groups', 'cluster_known_vars', 'all', 'all_no_filter']

    parser = argparse.ArgumentParser(
        description = 'Make a summary of ARIBA report files, and Phandango files',
        usage = 'ariba summary [options] <outprefix> [report1.tsv report2.tsv ...]',
        epilog = 'Files must be listed after the output file and/or the option --fofn must be used. If both used, all files in the filename specified by --fofn AND the files listed after the output file will be used as input.')
    parser.add_argument('-f', '--fofn', help='File of filenames of ariba reports in tsv format (not xls) to be summarised. Must be used if no input files listed after the outfile.', metavar='FILENAME')
    parser.add_argument('--preset', choices=presets, help='Shorthand for setting --cluster_cols,--col_filter,--row_filter,--known_vars,--novel_vars. Using this overrides those options', metavar='|'.join(presets))
    parser.add_argument('--cluster_cols', help='Comma separated list of cluster columns to include. Choose from: assembled, has_res, ref_seq, pct_id, known_var, novel_var [%(default)s]', default='has_res', metavar='col1,col2,...')
    parser.add_argument('--col_filter', choices=['y', 'n'], default='y', help='Choose whether columns where all values are "no" or "NA" are removed [%(default)s]', metavar='y|n')
    parser.add_argument('--row_filter', choices=['y', 'n'], default='y', help='Choose whether rows where all values are "no" or "NA" are removed [%(default)s]', metavar='y|n')
    parser.add_argument('--var_cols', help='Comma separated list of variant columns to include. Choose from: groups, grouped, ungrouped, novel [none by default]', metavar='col1,col2,...', default='')
    parser.add_argument('--min_id', type=float, help='Minimum percent identity cutoff to count as assembled [%(default)s]', default=90, metavar='FLOAT')
    parser.add_argument('--verbose', action='store_true', help='Be verbose')
    parser.add_argument('outprefix', help='Prefix of output files')
    parser.add_argument('infiles', nargs='*', help='Files to be summarised')
    options = parser.parse_args()
    if len(options.infiles) == 0:
        options.infiles = None

    options = use_preset(options)

    s = ariba.summary.Summary(
        options.outprefix,
        fofn=options.fofn,
        filenames=options.infiles,
        filter_rows=options.col_filter == 'y',
        filter_columns=options.row_filter == 'y',
        min_id=options.min_id,
        cluster_cols=options.cluster_cols,
        variant_cols=options.var_cols,
        verbose=options.verbose
    )
    s.run()



import signal
import time
import atexit
import os
import copy
import tempfile
import pickle
import itertools
import sys
import shutil
import openpyxl
import multiprocessing
import pysam
import pyfastaq
from ariba import cluster, common, mapping, histogram, read_store, report, report_filter, reference_data

class Error (Exception): pass


def _run_cluster(obj, verbose, clean, fails_dir):
    failed_clusters = os.listdir(fails_dir)

    if len(failed_clusters) > 0:
        print('Other clusters failed. Will not start cluster', obj.name, file=sys.stderr)
        return obj

    if verbose:
        print('Start running cluster', obj.name, 'in directory', obj.root_dir, flush=True)
    try:
        obj.run()
    except:
        print('Failed cluster:', obj.name, file=sys.stderr)
        with open(os.path.join(fails_dir, obj.name), 'w'):
            pass

    if verbose:
        print('Finished running cluster', obj.name, 'in directory', obj.root_dir, flush=True)

    if clean:
        if verbose:
            print('Deleting cluster dir', obj.root_dir, flush=True)
        if os.path.exists(obj.root_dir):
            shutil.rmtree(obj.root_dir)

    return obj


class Clusters:
    def __init__(self,
      refdata_dir,
      reads_1,
      reads_2,
      outdir,
      extern_progs,
      version_report_lines=None,
      assembly_kmer=21,
      assembly_coverage=100,
      threads=1,
      verbose=False,
      assembler='spades',
      spades_other=None,
      max_insert=1000,
      min_scaff_depth=10,
      nucmer_min_id=90,
      nucmer_min_len=20,
      nucmer_breaklen=200,
      assembled_threshold=0.95,
      unique_threshold=0.03,
      max_gene_nt_extend=30,
      bowtie2_preset='very-sensitive-local',
      clean=True,
      tmp_dir=None,
    ):
        self.refdata_dir = os.path.abspath(refdata_dir)
        self.refdata, self.cluster_ids = self._load_reference_data_from_dir(refdata_dir)
        self.reads_1 = os.path.abspath(reads_1)
        self.reads_2 = os.path.abspath(reads_2)
        self.outdir = os.path.abspath(outdir)
        self.extern_progs = extern_progs

        if version_report_lines is None:
            self.version_report_lines = []
        else:
            self.version_report_lines = version_report_lines

        self.clean = clean
        self.logs_dir = os.path.join(self.outdir, 'Logs')

        self.assembler = assembler
        assert self.assembler in ['spades']
        self.assembly_kmer = assembly_kmer
        self.assembly_coverage = assembly_coverage
        self.spades_other = spades_other

        self.cdhit_files_prefix = os.path.join(self.refdata_dir, 'cdhit')
        self.cdhit_cluster_representatives_fa = self.cdhit_files_prefix + '.cluster_representatives.fa'
        self.bam_prefix = os.path.join(self.outdir, 'map_reads_to_cluster_reps')
        self.bam = self.bam_prefix + '.bam'
        self.report_file_all_tsv = os.path.join(self.outdir, 'report.all.tsv')
        self.report_file_all_xls = os.path.join(self.outdir, 'report.all.xls')
        self.report_file_filtered_prefix = os.path.join(self.outdir, 'report')
        self.catted_assembled_seqs_fasta = os.path.join(self.outdir, 'assembled_seqs.fa.gz')
        self.catted_genes_matching_refs_fasta = os.path.join(self.outdir, 'assembled_genes.fa.gz')
        self.threads = threads
        self.verbose = verbose

        self.max_insert = max_insert
        self.bowtie2_preset = bowtie2_preset

        self.insert_hist_bin = 10
        self.insert_hist = histogram.Histogram(self.insert_hist_bin)
        self.insert_size = None
        self.insert_sspace_sd = None
        self.insert_proper_pair_max = None

        self.min_scaff_depth = min_scaff_depth
        self.nucmer_min_id = nucmer_min_id
        self.nucmer_min_len = nucmer_min_len
        self.nucmer_breaklen = nucmer_breaklen

        self.assembled_threshold = assembled_threshold
        self.unique_threshold = unique_threshold
        self.max_gene_nt_extend = max_gene_nt_extend

        self.cluster_to_dir = {}  # gene name -> abs path of cluster directory
        self.clusters = {}        # gene name -> Cluster object
        self.cluster_read_counts = {} # gene name -> number of reads
        self.cluster_base_counts = {} # gene name -> number of bases
        self.pool = None
        self.fails_dir = os.path.join(self.outdir ,'.fails')
        self.clusters_all_ran_ok = True

        for d in [self.outdir, self.logs_dir, self.fails_dir]:
            try:
                os.mkdir(d)
            except:
                raise Error('Error mkdir ' + d)

        if tmp_dir is None:
            if 'ARIBA_TMPDIR' in os.environ:
                tmp_dir = os.path.abspath(os.environ['ARIBA_TMPDIR'])
            else:
                tmp_dir = self.outdir

        if not os.path.exists(tmp_dir):
            raise Error('Temporary directory ' + tmp_dir + ' not found. Cannot continue')

        if self.clean:
            self.tmp_dir_obj = tempfile.TemporaryDirectory(prefix='ariba.tmp.', dir=os.path.abspath(tmp_dir))
            self.tmp_dir = self.tmp_dir_obj.name
        else:
            self.tmp_dir_obj = None
            self.tmp_dir = os.path.join(self.outdir, 'clusters')
            try:
                os.mkdir(self.tmp_dir)
            except:
                raise Error('Error making directory ' + self.tmp_dir)

        if self.verbose:
            print('Temporary directory:', self.tmp_dir)

        for i in [x for x in dir(signal) if x.startswith("SIG") and x not in {'SIGCHLD', 'SIGCLD'}]:
            try:
                signum = getattr(signal, i)
                signal.signal(signum, self._receive_signal)
            except:
                pass


    def _stop_pool(self):
        if self.pool is None:
            return
        self.pool.close()
        self.pool.terminate()
        while len(multiprocessing.active_children()) > 0:
            time.sleep(1)


    def _emergency_stop(self):
        self._stop_pool()
        if self.clean:
            try:
                self.tmp_dir_obj.cleanup()
            except:
                pass


    def _receive_signal(self, signum, stack):
        print('Stopping! Signal received:', signum, file=sys.stderr, flush=True)
        self._emergency_stop()
        sys.exit(1)


    @classmethod
    def _load_reference_data_info_file(cls, filename):
        data = {
            'genetic_code': None
        }

        with open(filename) as f:
            for line in f:
                key, val = line.rstrip().split('\t')
                if key in data:
                    data[key] = val

        if None in data.values():
            missing_values = [x for x in data if data[x] is None]
            raise Error('Error reading reference info file ' + filename + '. These values not found: ' + ','.join(missing_values))

        data['genetic_code'] = int(data['genetic_code'])
        return data


    @staticmethod
    def _load_reference_data_from_dir(indir):
        if not os.path.exists(indir):
            raise Error('Error loading reference data. Input directory ' + indir + ' not found. Cannot continue')

        variants_only_fa = os.path.join(indir, 'refcheck.01.check_variants.variants_only.fa')
        presence_absence_fa = os.path.join(indir, 'refcheck.01.check_variants.presence_absence.fa')
        non_coding_fa = os.path.join(indir, 'refcheck.01.check_variants.non_coding.fa')
        metadata_tsv = os.path.join(indir, 'refcheck.01.check_variants.tsv')
        info_file = os.path.join(indir, 'info.txt')
        clusters_file = os.path.join(indir, 'cdhit.clusters.pickle')
        params = Clusters._load_reference_data_info_file(info_file)
        refdata = reference_data.ReferenceData(
            presence_absence_fa=presence_absence_fa if os.path.exists(presence_absence_fa) else None,
            variants_only_fa=variants_only_fa if os.path.exists(variants_only_fa) else None,
            non_coding_fa=non_coding_fa if os.path.exists(non_coding_fa) else None,
            metadata_tsv=metadata_tsv if os.path.exists(metadata_tsv) else None,
            genetic_code=params['genetic_code'],
        )

        with open(clusters_file, 'rb') as f:
            cluster_ids = pickle.load(f)

        return refdata, cluster_ids


    def _map_reads_to_clustered_genes(self):
        mapping.run_bowtie2(
            self.reads_1,
            self.reads_2,
            self.cdhit_cluster_representatives_fa,
            self.bam_prefix,
            threads=self.threads,
            samtools=self.extern_progs.exe('samtools'),
            bowtie2=self.extern_progs.exe('bowtie2'),
            bowtie2_preset=self.bowtie2_preset,
            verbose=self.verbose,
            remove_both_unmapped=True,
        )


    def _bam_to_clusters_reads(self):
        '''Sets up ReadStore of reads for all the clusters. Also gathers histogram data of insert size'''
        reads_file_for_read_store = os.path.join(self.outdir, 'reads')
        f_out = pyfastaq.utils.open_file_write(reads_file_for_read_store)

        sam_reader = pysam.Samfile(self.bam, "rb")
        sam1 = None
        self.proper_pairs = 0

        for s in sam_reader.fetch(until_eof=True):
            if sam1 is None:
                sam1 = s
                continue

            ref_seqs = set()
            if not s.is_unmapped:
                ref_seqs.add(sam_reader.getrname(s.tid))
            if not sam1.is_unmapped:
                ref_seqs.add(sam_reader.getrname(sam1.tid))

            read1 = mapping.sam_to_fastq(sam1)
            read2 = mapping.sam_to_fastq(s)
            if read1.id.endswith('/2'):
                read1, read2 = read2, read1

            insert = mapping.sam_pair_to_insert(s, sam1)
            if insert is not None:
                self.insert_hist.add(insert)
                self.proper_pairs += 1

            for ref in ref_seqs:
                if ref not in self.cluster_to_dir:
                    new_dir = os.path.join(self.tmp_dir, ref)
                    self.cluster_to_dir[ref] = new_dir
                    if self.verbose:
                        print('New cluster with reads that hit:', ref, flush=True)

                self.cluster_read_counts[ref] = self.cluster_read_counts.get(ref, 0) + 2
                self.cluster_base_counts[ref] = self.cluster_base_counts.get(ref, 0) + len(read1) + len(read2)
                print(ref, self.cluster_read_counts[ref] - 1, read1.seq, read1.qual, sep='\t', file=f_out)
                print(ref, self.cluster_read_counts[ref], read2.seq, read2.qual, sep='\t', file=f_out)

            sam1 = None

        pyfastaq.utils.close(f_out)

        if len(self.cluster_read_counts):
            if self.verbose:
                filehandle = sys.stdout
            else:
                filehandle = None

            self.read_store = read_store.ReadStore(
              reads_file_for_read_store,
              os.path.join(self.outdir, 'read_store'),
              log_fh=filehandle
            )

        os.unlink(reads_file_for_read_store)

        if self.verbose:
            print('Found', self.proper_pairs, 'proper read pairs')
            print('Total clusters to perform local assemblies:', len(self.cluster_to_dir), flush=True)


    def _set_insert_size_data(self):
        if len(self.insert_hist) == 0:
            return False
        else:
            (x, self.insert_size, pc95, self.insert_sspace_sd) = self.insert_hist.stats()
            self.insert_sspace_sd = min(1, self.insert_sspace_sd)
            self.insert_proper_pair_max = 1.1 * pc95
            if self.verbose:
                print('\nInsert size information from reads mapped to reference genes:')
                print('Insert size:', self.insert_size, sep='\t')
                print('Insert sspace sd:', self.insert_sspace_sd, sep='\t')
                print('Max insert:', self.insert_proper_pair_max, sep='\t')
                print()
            return True


    def _init_and_run_clusters(self):
        if len(self.cluster_to_dir) == 0:
            raise Error('Did not get any reads mapped to genes. Cannot continue')

        counter = 0
        cluster_list = []
        self.log_files = []

        for seq_type in sorted(self.cluster_ids):
            if self.cluster_ids[seq_type] is None:
                continue

            for seq_name in sorted(self.cluster_ids[seq_type]):
                if seq_name not in self.cluster_to_dir:
                    continue
                counter += 1
                if self.verbose:
                    print('Constructing cluster', seq_name + '.', counter, 'of', str(len(self.cluster_to_dir)))
                new_dir = self.cluster_to_dir[seq_name]
                self.log_files.append(os.path.join(self.logs_dir, seq_name + '.log'))

                cluster_list.append(cluster.Cluster(
                    new_dir,
                    seq_name,
                    self.refdata,
                    self.cluster_read_counts[seq_name],
                    self.cluster_base_counts[seq_name],
                    fail_file=os.path.join(self.fails_dir, seq_name),
                    read_store=self.read_store,
                    reference_names=self.cluster_ids[seq_type][seq_name],
                    logfile=self.log_files[-1],
                    assembly_coverage=self.assembly_coverage,
                    assembly_kmer=self.assembly_kmer,
                    assembler=self.assembler,
                    max_insert=self.insert_proper_pair_max,
                    min_scaff_depth=self.min_scaff_depth,
                    nucmer_min_id=self.nucmer_min_id,
                    nucmer_min_len=self.nucmer_min_len,
                    nucmer_breaklen=self.nucmer_breaklen,
                    reads_insert=self.insert_size,
                    sspace_k=self.min_scaff_depth,
                    sspace_sd=self.insert_sspace_sd,
                    threads=1, # clusters now run in parallel, so this should always be 1!
                    bcf_min_dp=10,            # let the user change this in a future version?
                    bcf_min_dv=5,             # let the user change this in a future version?
                    bcf_min_dv_over_dp=0.3,   # let the user change this in a future version?
                    bcf_min_qual=20,          # let the user change this in a future version?
                    assembled_threshold=self.assembled_threshold,
                    unique_threshold=self.unique_threshold,
                    max_gene_nt_extend=self.max_gene_nt_extend,
                    bowtie2_preset=self.bowtie2_preset,
                    spades_other_options=self.spades_other,
                    clean=self.clean,
                    extern_progs=self.extern_progs,
                ))

        try:
            if self.threads > 1:
                self.pool = multiprocessing.Pool(self.threads)
                cluster_list = self.pool.starmap(_run_cluster, zip(cluster_list, itertools.repeat(self.verbose), itertools.repeat(self.clean), itertools.repeat(self.fails_dir)))
            else:
                for c in cluster_list:
                    _run_cluster(c, self.verbose, self.clean, self.fails_dir)
        except:
            self.clusters_all_ran_ok = False

        if len(os.listdir(self.fails_dir)) > 0:
            self.clusters_all_ran_ok = False

        self.clusters = {c.name: c for c in cluster_list}


    @staticmethod
    def _write_reports(clusters_in, tsv_out, xls_out=None):
        columns = copy.copy(report.columns)
        columns[0] = '#' + columns[0]

        f = pyfastaq.utils.open_file_write(tsv_out)
        print('\t'.join(columns), file=f)

        columns[0] = columns[0][1:]
        workbook = openpyxl.Workbook()
        worksheet = workbook.worksheets[0]
        worksheet.title = 'ARIBA_report'
        worksheet.append(columns)

        for seq_name in sorted(clusters_in):
            if clusters_in[seq_name].report_lines is None:
                continue

            for line in clusters_in[seq_name].report_lines:
                print(line, file=f)
                worksheet.append(line.split('\t'))

        pyfastaq.utils.close(f)
        if xls_out is not None:
            workbook.save(xls_out)


    def _write_catted_assembled_seqs_fasta(self, outfile):
        f = pyfastaq.utils.open_file_write(outfile)

        for gene in sorted(self.clusters):
            try:
                seq_dict = self.clusters[gene].assembly_compare.assembled_reference_sequences
            except:
                continue

            for seq_name in sorted(seq_dict):
                print(seq_dict[seq_name], file=f)

        pyfastaq.utils.close(f)


    def _write_catted_genes_matching_refs_fasta(self, outfile):
        f = pyfastaq.utils.open_file_write(outfile)

        for gene in sorted(self.clusters):
            if self.clusters[gene].assembly_compare is not None and self.clusters[gene].assembly_compare.gene_matching_ref is not None:
                seq = copy.copy(self.clusters[gene].assembly_compare.gene_matching_ref)
                seq.id += '.' + '.'.join([
                    self.clusters[gene].assembly_compare.gene_matching_ref_type,
                    str(self.clusters[gene].assembly_compare.gene_start_bases_added),
                    str(self.clusters[gene].assembly_compare.gene_end_bases_added)
                ])
                print(seq, file=f)

        pyfastaq.utils.close(f)


    def _clean(self):
        if self.clean:
            shutil.rmtree(self.fails_dir)

            try:
                self.tmp_dir_obj.cleanup()
            except:
                pass

            if self.verbose:
                print('Deleting Logs directory', self.logs_dir)
            try:
                shutil.rmtree(self.logs_dir)
            except:
                pass

            if self.verbose:
                print('Deleting reads store files', self.read_store.outfile + '[.tbi]')
            try:
                self.read_store.clean()
            except:
                pass
        else:
            if self.verbose:
                print('Not deleting anything because --noclean used')


    def write_versions_file(self, original_dir):
        with open('version_info.txt', 'w') as f:
            print('ARIBA run with this command:', file=f)
            print(' '.join([sys.argv[0]] + ['run'] + sys.argv[1:]), file=f)
            print('from this directory:', original_dir, file=f)
            print(file=f)
            print(*self.version_report_lines, sep='\n', file=f)


    def run(self):
        try:
            self._run()
        except Error as err:
            self._emergency_stop()
            raise Error('Something went wrong during ariba run. Cannot continue. Error was:\n' + str(err))


    def _run(self):
        cwd = os.getcwd()
        os.chdir(self.outdir)
        self.write_versions_file(cwd)

        if self.verbose:
            print('{:_^79}'.format(' Mapping reads to clustered genes '), flush=True)
        self._map_reads_to_clustered_genes()

        if self.verbose:
            print('Finished mapping\n')
            print('{:_^79}'.format(' Generating clusters '), flush=True)
        self._bam_to_clusters_reads()
        if self.clean:
            if self.verbose:
                print('Deleting BAM', self.bam, flush=True)
            os.unlink(self.bam)

        if len(self.cluster_to_dir) > 0:
            got_insert_data_ok = self._set_insert_size_data()
            if not got_insert_data_ok:
                print('WARNING: not enough proper read pairs (found ' + str(self.proper_pairs) + ') to determine insert size.', file=sys.stderr)
                print('This probably means that very few reads were mapped at all. No local assemblies will be run', file=sys.stderr)
                if self.verbose:
                    print('Not enough proper read pairs mapped to determine insert size. Skipping all assemblies.', flush=True)
            else:
                if self.verbose:
                    print('{:_^79}'.format(' Assembling each cluster '))
                    print('Will run', self.threads, 'cluster(s) in parallel', flush=True)
                self._init_and_run_clusters()
                if self.verbose:
                    print('Finished assembling clusters\n')
        else:
            if self.verbose:
                print('No reads mapped. Skipping all assemblies', flush=True)
            print('WARNING: no reads mapped to reference genes. Therefore no local assemblies will be run', file=sys.stderr)

        if not self.clusters_all_ran_ok:
            raise Error('At least one cluster failed! Stopping...')

        if self.verbose:
            print('{:_^79}'.format(' Writing reports '), flush=True)
            print('Making', self.report_file_all_tsv)
        self._write_reports(self.clusters, self.report_file_all_tsv)

        if self.verbose:
            print('Making', self.report_file_filtered_prefix + '.tsv')
        rf = report_filter.ReportFilter(infile=self.report_file_all_tsv)
        rf.run(self.report_file_filtered_prefix)

        if self.verbose:
            print()
            print('{:_^79}'.format(' Writing fasta of assembled sequences '), flush=True)
            print(self.catted_assembled_seqs_fasta, 'and', self.catted_genes_matching_refs_fasta, flush=True)
        self._write_catted_assembled_seqs_fasta(self.catted_assembled_seqs_fasta)
        self._write_catted_genes_matching_refs_fasta(self.catted_genes_matching_refs_fasta)

        clusters_log_file = os.path.join(self.outdir, 'log.clusters.gz')
        if self.verbose:
            print()
            print('{:_^79}'.format(' Catting cluster log files '), flush=True)
            print('Writing file', clusters_log_file, flush=True)
        common.cat_files(self.log_files, clusters_log_file)

        if self.verbose:
            print()
            print('{:_^79}'.format(' Cleaning files '), flush=True)
        self._clean()

        if self.clusters_all_ran_ok and self.verbose:
            print('\nAll done!\n')

        os.chdir(cwd)



import argparse
import os
import sys
import ariba


def run():
    parser = argparse.ArgumentParser(
        description = 'ARIBA: Antibiotic Resistance Identification By Assembly',
        usage = 'ariba run [options] <prepareref_dir> <reads1.fq> <reads2.fq> <outdir>')
    parser.add_argument('prepareref_dir', help='Name of output directory when "ariba prepareref" was run')
    parser.add_argument('reads_1', help='Name of fwd reads fastq file')
    parser.add_argument('reads_2', help='Name of rev reads fastq file')
    parser.add_argument('outdir', help='Output directory (must not already exist)')

    nucmer_group = parser.add_argument_group('nucmer options')
    nucmer_group.add_argument('--nucmer_min_id', type=int, help='Minimum alignment identity (delta-filter -i) [%(default)s]', default=90, metavar='INT')
    nucmer_group.add_argument('--nucmer_min_len', type=int, help='Minimum alignment length (delta-filter -i) [%(default)s]', default=20, metavar='INT')
    nucmer_group.add_argument('--nucmer_breaklen', type=int, help='Value to use for -breaklen when running nucmer [%(default)s]', default=200, metavar='INT')

    assembly_group = parser.add_argument_group('Assembly options')
    assembly_group.add_argument('--assembly_cov', type=int, help='Target read coverage when sampling reads for assembly [%(default)s]', default=50, metavar='INT')
    assembly_group.add_argument('--assembler_k', type=int, help='kmer size to use with assembler. You can use 0 to set kmer to 2/3 of the read length. Warning - lower kmers are usually better. [%(default)s]', metavar='INT', default=21)
    assembly_group.add_argument('--spades_other', help='Put options string to be used with spades in quotes. This will NOT be sanity checked. Do not use -k (see --assembler_k), --untrusted-contigs (it is always used), or -t [%(default)s]', default="--only-assembler -m 4", metavar="OPTIONS")
    assembly_group.add_argument('--min_scaff_depth', type=int, help='Minimum number of read pairs needed as evidence for scaffold link between two contigs. This is also the value used for sspace -k when scaffolding [%(default)s]', default=10, metavar='INT')

    other_group = parser.add_argument_group('Other options')
    other_group.add_argument('--threads', type=int, help='Number of threads [%(default)s]', default=1, metavar='INT')
    bowtie2_presets = ['very-fast-local', 'fast-local', 'sensitive-local', 'very-sensitive-local']
    other_group.add_argument('--bowtie2_preset', choices=bowtie2_presets, help='Preset option for bowtie2 mapping [%(default)s]', default='very-sensitive-local', metavar='|'.join(bowtie2_presets))
    other_group.add_argument('--assembled_threshold', type=float, help='If proportion of gene assembled (regardless of into how many contigs) is at least this value then the flag gene_assembled is set [%(default)s]', default=0.95, metavar='FLOAT (between 0 and 1)')
    other_group.add_argument('--gene_nt_extend', type=int, help='Max number of nucleotides to extend ends of gene matches to look for start/stop codons [%(default)s]', default=30, metavar='INT')
    other_group.add_argument('--unique_threshold', type=float, help='If proportion of bases in gene assembled more than once is <= this value, then the flag unique_contig is set [%(default)s]', default=0.03, metavar='FLOAT (between 0 and 1)')
    other_group.add_argument('--noclean', action='store_true', help='Do not clean up intermediate files')
    other_group.add_argument('--tmp_dir', help='Existing directory in which to create a temporary directory used for local assemblies')
    other_group.add_argument('--verbose', action='store_true', help='Be verbose')

    options = parser.parse_args()

    reads_not_found = []

    for filename in [options.reads_1, options.reads_2]:
        if not os.path.exists(filename):
            reads_not_found.append(filename)
        elif options.verbose:
            print('Found reads file:', filename)

    if len(reads_not_found):
        print('\nThe following reads file(s) were not found:', file=sys.stderr)
        print(*reads_not_found, sep='\n', file=sys.stderr)
        print('Cannot continue', file=sys.stderr)
        sys.exit(1)

    if not os.path.exists(options.prepareref_dir):
        print('Input directory', options.prepareref_dir, 'not found. Cannot continue', file=sys.stderr)
        sys.exit(1)

    extern_progs, version_report_lines = ariba.versions.get_all_versions()
    if options.verbose:
        print(*version_report_lines, sep='\n')

    c = ariba.clusters.Clusters(
          options.prepareref_dir,
          options.reads_1,
          options.reads_2,
          options.outdir,
          extern_progs,
          version_report_lines=version_report_lines,
          assembly_kmer=options.assembler_k,
          assembly_coverage=options.assembly_cov,
          assembler='spades',
          threads=options.threads,
          verbose=options.verbose,
          min_scaff_depth=options.min_scaff_depth,
          nucmer_min_id=options.nucmer_min_id,
          nucmer_min_len=options.nucmer_min_len,
          nucmer_breaklen=options.nucmer_breaklen,
          spades_other=options.spades_other,
          assembled_threshold=options.assembled_threshold,
          unique_threshold=options.unique_threshold,
          max_gene_nt_extend=options.gene_nt_extend,
          bowtie2_preset=options.bowtie2_preset,
          clean=(not options.noclean),
          tmp_dir=options.tmp_dir,
        )
    c.run()




import argparse
from ariba import aln_to_metadata


def run():
    coding_choices = ['coding', 'noncoding']
    parser = argparse.ArgumentParser(
        description = 'Converts multi-alignment fasta and SNP info to metadata',
        usage = 'ariba aln2meta [options] <aln_fasta> <variants_tsv> <(non)coding> <cluster_rep> <outprefix>'
    )

    parser.add_argument('--genetic_code', type=int, help='Number of genetic code to use. Currently supported 1,4,11 [%(default)s]', choices=[1,4,11], default=11, metavar='INT')
    parser.add_argument('aln_fasta', help='Multi-fasta file of alignments')
    parser.add_argument('variants_tsv', help='TSV file of variants information')
    parser.add_argument('coding_or_non', help='Sequences are coding or noncoding. Must be one of: ' + ' '.join(coding_choices), choices=coding_choices, metavar='(non)coding')
    parser.add_argument('cluster_rep', help='Name of sequence to be used as cluster representative. Must exactly match a sequence in aln_fasta file')
    parser.add_argument('outprefix', help='Prefix of output filenames')
    options = parser.parse_args()

    aln_to_meta = aln_to_metadata.AlnToMetadata(
      options.aln_fasta,
      options.variants_tsv,
      options.coding_or_non == 'coding',
      options.cluster_rep,
      genetic_code=options.genetic_code
    )
    aln_to_meta.run(options.outprefix)




import os
import copy
import re
import sys
import openpyxl
import pyfastaq
from ariba import flag, common, report, summary_cluster, summary_sample

class Error (Exception): pass

required_keys_for_difference = {'no', 'yes', 'yes_nonunique', 'fragmented'}

class Summary:
    def __init__(
      self,
      outprefix,
      filenames=None,
      fofn=None,
      filter_rows=True,
      filter_columns=True,
      min_id=90.0,
      cluster_cols='assembled,has_res,ref_seq,pct_id,known_var,novel_var',
      variant_cols='groups,grouped,ungrouped,novel',
      verbose=False,
    ):
        if filenames is None and fofn is None:
            raise Error('Error! Must supply filenames or fofn to Summary(). Cannot continue')

        if filenames is None:
            self.filenames = []
        else:
            self.filenames = filenames

        if fofn is not None:
            self.filenames.extend(self._load_fofn(fofn))

        self.cluster_columns = self._determine_cluster_cols(cluster_cols)
        self.var_columns = self._determine_var_cols(variant_cols)
        self.filter_rows = filter_rows
        self.filter_columns = filter_columns
        self.min_id = min_id
        self.outprefix = outprefix
        self.verbose = verbose


    @classmethod
    def _determine_cols(cls, cols_string, allowed_cols, error_string):
        if cols_string == '' or cols_string is None:
            return {x: False for x in allowed_cols}
        wanted_cols = set(cols_string.split(','))
        if not wanted_cols.issubset(allowed_cols):
            raise Error('Error in ' + error_string + '. Allowed values are: ' + str(','.join(list(allowed_cols))) + '. Got: ' + cols_string)
        return {x: x in wanted_cols for x in allowed_cols}


    @staticmethod
    def _determine_cluster_cols(cols_string):
        allowed_cols = {'assembled', 'has_res', 'ref_seq', 'pct_id', 'known_var', 'novel_var'}
        return Summary._determine_cols(cols_string, allowed_cols, 'cluster columns')


    @staticmethod
    def _determine_var_cols(cols_string):
        allowed_cols = {'groups', 'grouped', 'ungrouped', 'novel'}
        return Summary._determine_cols(cols_string, allowed_cols, 'variant columns')


    def _load_fofn(self, fofn):
        f = pyfastaq.utils.open_file_read(fofn)
        filenames = [x.rstrip() for x in f.readlines()]
        pyfastaq.utils.close(f)
        return filenames


    def _check_files_exist(self):
        for fname in self.filenames:
            if not os.path.exists(fname):
                raise Error('File not found: "' + fname + '". Cannot continue')


    @classmethod
    def _load_input_files(cls, filenames, min_id, verbose=False):
        samples = {}
        for filename in filenames:
            samples[filename] = summary_sample.SummarySample(filename, min_pc_id=min_id)
            samples[filename].run()
            if verbose:
                print('Loaded file', filename, flush=True)
        return samples


    @classmethod
    def _get_all_cluster_names(cls, samples_dict):
        '''Input should be output of _load_input_files'''
        cluster_names = set()
        for filename, sample in samples_dict.items():
            cluster_names.update(set(sample.clusters.keys()))
        return cluster_names


    @classmethod
    def _get_all_variant_columns(cls, samples_dict):
        '''Input should be output of _load_input_files'''
        columns = {}
        for filename, sample in samples_dict.items():
            for cluster in sample.column_summary_data:
                if sample.column_summary_data[cluster]['assembled'] == 'yes':
                    for key, tuple_set in sample.variant_column_names_tuples.items():
                        for t in tuple_set:
                            if key not in columns:
                                columns[key] = set()
                            columns[key].add(t)
        return columns


    @classmethod
    def _get_all_var_groups(cls, samples_dict):
        groups = {}
        for filename, sample in samples_dict.items():
            for name, name_set in sample.var_groups.items():
                if name not in groups:
                    groups[name] = set()
                groups[name].update(name_set)
        return groups


    def _gather_output_rows(self):
        all_cluster_names = Summary._get_all_cluster_names(self.samples)
        all_var_columns = Summary._get_all_variant_columns(self.samples)
        if self.var_columns['groups']:
            var_groups = Summary._get_all_var_groups(self.samples)
        else:
            var_groups = set()
        rows = {}

        for filename, sample in self.samples.items():
            rows[filename] = {}

            for cluster in all_cluster_names:
                rows[filename][cluster] = {}

                if cluster in sample.column_summary_data and sample.column_summary_data[cluster]['assembled'].startswith('yes'):
                    rows[filename][cluster] = sample.column_summary_data[cluster]
                else:
                    rows[filename][cluster] = {
                        'assembled': 'no',
                        'has_res': 'no',
                        'ref_seq': 'NA',
                        'known_var': 'NA',
                        'novel_var': 'NA',
                        'pct_id': 'NA'
                    }

                if self.var_columns['groups']:
                    for group_name in var_groups[cluster]:
                        if cluster in sample.var_groups and group_name in sample.var_groups[cluster]:
                            rows[filename][cluster]['vgroup.' + group_name] = 'yes'
                        else:
                            rows[filename][cluster]['vgroup.' + group_name] = 'no'

                if cluster in all_var_columns:
                    for (ref_name, variant, grouped_or_novel, group_name) in all_var_columns[cluster]:
                        if not self.var_columns[grouped_or_novel]:
                            continue

                        key = ref_name + '.' + variant
                        if rows[filename][cluster]['assembled'] == 'no':
                            rows[filename][cluster][key] = 'NA'
                        elif cluster in sample.variant_column_names_tuples and (ref_name, variant, grouped_or_novel, group_name) in sample.variant_column_names_tuples[cluster]:
                            rows[filename][cluster][key] = 'yes'
                        else:
                            rows[filename][cluster][key] = 'no'

                for key, wanted in self.cluster_columns.items():
                    if not wanted:
                        del rows[filename][cluster][key]

        return rows


    @classmethod
    def _to_matrix(cls, filenames, rows, cluster_cols):
        '''rows = output from _gather_output_rows().
           filenames = self.filenames
           cluster_cols = self.cluster_columns'''
        matrix = []
        making_header_lines = True
        phandango_header = ['name']
        phandago_suffixes = {'assembled': ':o1', 'has_res': ':o1', 'ref_seq': ':o2', 'pct_id': ':c1', 'known_var': ':o1', 'novel_var': 'o1'}
        csv_header = ['name']
        all_cluster_cols_in_order = ['assembled', 'has_res', 'ref_seq', 'pct_id', 'known_var', 'novel_var']
        all_cluster_cols_in_order_set = set(['assembled', 'has_res', 'ref_seq', 'pct_id', 'known_var', 'novel_var'])
        cluster_cols_in_order = [x for x in all_cluster_cols_in_order if cluster_cols[x]]
        cluster_cols_set = set(cluster_cols_in_order)

        for filename in filenames:
            assert filename in rows
            line = [filename]

            for cluster_name in sorted(rows[filename]):
                for col in cluster_cols_in_order:
                    if making_header_lines:
                        csv_header.append(cluster_name + '.' + col)
                        phandango_header.append(cluster_name + '.' + col + '.' + phandago_suffixes[col])

                    line.append(rows[filename][cluster_name][col])

                for col in sorted(rows[filename][cluster_name]):
                    if col in all_cluster_cols_in_order_set:
                        continue

                    if making_header_lines:
                        csv_header.append(cluster_name + '.' + col)
                        phandango_header.append(cluster_name + '.' + col + ':o1')

                    line.append(rows[filename][cluster_name][col])

            making_header_lines = False
            matrix.append(line)

        return phandango_header, csv_header, matrix


    @classmethod
    def _filter_matrix_rows(cls, matrix):
        '''matrix = output from _to_matrix'''
        indexes_to_keep = []

        for i in range(len(matrix)):
            keep_row = False
            for element in matrix[i]:
                if element not in {'NA', 'no'}:
                    keep_row = True
                    break
            if keep_row:
                indexes_to_keep.append(i)

        return [matrix[i] for i in indexes_to_keep]


    @classmethod
    def _filter_matrix_columns(cls, matrix, phandango_header, csv_header):
        '''phandango_header, csv_header, matrix = output from _to_matrix'''
        indexes_to_keep = set()

        for row in matrix:
            for i in range(len(row)):
                if row[i] not in {'NA', 'no'}:
                    indexes_to_keep.add(i)

        indexes_to_keep = sorted(list(indexes_to_keep))

        for i in range(len(matrix)):
            matrix[i] = [matrix[i][j] for j in indexes_to_keep]

        phandango_header = [phandango_header[i] for i in indexes_to_keep]
        csv_header = [csv_header[i] for i in indexes_to_keep]
        return phandango_header, csv_header, matrix


    @classmethod
    def _add_phandango_colour_columns(cls, header, matrix):
        header = copy.deepcopy(header)
        matrix = copy.deepcopy(matrix)
        cols_to_add_colour_col = [i for i in range(len(header)) if header[i].endswith(':o1')]
        field_to_col = {
            'yes': '#1f78b4',
            'yes_nonunique': '#a6cee3',
            'no': '#33a02c',
            'NA': '#b2df8a',
        }

        cols_to_add_colour_col.reverse()

        for col_index in cols_to_add_colour_col:
            header[col_index] = header[col_index][:-3]
            header.insert(col_index + 1, header[col_index] + ':colour')

            for row_index in range(len(matrix)):
                colour = field_to_col[matrix[row_index][col_index]]
                matrix[row_index].insert(col_index + 1, colour)

        return header, matrix


    @classmethod
    def _matrix_to_csv(cls, matrix, header, outfile):
        f = pyfastaq.utils.open_file_write(outfile)
        print(*header, sep=',', file=f)
        for line in matrix:
            print(*line, sep=',', file=f)
        pyfastaq.utils.close(f)


    @staticmethod
    def _distance_score_between_values(value1, value2):
        value_set = {value1, value2}
        if value_set.isdisjoint(required_keys_for_difference) or value1 == value2 or value_set == {'NA', 'no'}:
            return 0
        else:
            return 1


    @classmethod
    def _distance_score_between_lists(cls, scores1, scores2):
        assert len(scores1) == len(scores2)
        return sum([cls._distance_score_between_values(scores1[i], scores2[i]) for i in range(1, len(scores1))])


    @classmethod
    def _write_distance_matrix(cls, lines, outfile):
        if len(lines) < 2:
            raise Error('Cannot calculate distance matrix to make tree for phandango.\n' +
                        'Only one sample present.')

        if len(lines[0]) < 2:
            raise Error('Cannot calculate distance matrix to make tree for phandango. Not enough columns')

        scores = [[0 for i in range(len(lines))] for j in range(len(lines))]

        for i in range(len(lines)):
            for j in range(i + 1, len(lines), 1):
                scores[i][j] = Summary._distance_score_between_lists(lines[i], lines[j])
                scores[j][i] = scores[i][j]

        with open(outfile, 'w') as f:
            sample_names = [x[0] for x in lines]
            print(*sample_names, sep='\t', file=f)
            for i in range(len(scores)):
                print(lines[i][0], *scores[i][1:], sep='\t', file=f)


    @classmethod
    def _newick_from_dist_matrix(cls, distance_file, outfile):
        r_script = outfile + '.tmp.R'

        with open(r_script, 'w') as f:
            print('library(ape)', file=f)
            print('a=read.table("', distance_file, '", header=TRUE, row.names=1, comment.char="")', sep='', file=f)
            print('h=hclust(dist(a))', file=f)
            print('write.tree(as.phylo(h), file="', outfile, '")', sep='', file=f)

        common.syscall('Rscript --no-save ' + r_script)
        if os.path.exists(r_script + 'out'):
            os.unlink(r_script + 'out')
        os.unlink(r_script)


    def run(self):
        if self.verbose:
            print('Loading input files...', flush=True)
        self._check_files_exist()
        self.samples = self._load_input_files(self.filenames, self.min_id, verbose=self.verbose)
        if self.verbose:
            print('Generating output rows', flush=True)
        self.rows = self._gather_output_rows()
        phandango_header, csv_header, matrix = Summary._to_matrix(self.filenames, self.rows, self.cluster_columns)

        if self.filter_rows:
            if self.verbose:
                print('Filtering rows', flush=True)
            matrix = Summary._filter_matrix_rows(matrix)

        if len(matrix) == 0:
            print('No rows left after filtering rows. Cannot continue', file=sys.stderr)
            sys.exit(1)

        if self.filter_columns:
            if self.verbose:
                print('Filtering columns', flush=True)
            phandango_header, csv_header, matrix = Summary._filter_matrix_columns(matrix, phandango_header, csv_header)

        if len(matrix) == 0 or len(matrix[0]) == 0:
            print('No columns left after filtering columns. Cannot continue', file=sys.stderr)

        csv_file = self.outprefix + '.csv'
        if self.verbose:
            print('Writing csv file', csv_file, flush=True)
        Summary._matrix_to_csv(matrix, csv_header, csv_file)

        if len(matrix) > 1:
            if self.verbose:
                print('Making Phandango csv file', csv_file, flush=True)
            csv_file = self.outprefix + '.phandango.csv'
            phandango_header, phandango_matrix = Summary._add_phandango_colour_columns(phandango_header, matrix)
            Summary._matrix_to_csv(phandango_matrix, phandango_header, csv_file)
            dist_matrix_file = self.outprefix + '.phandango.distance_matrix'
            tree_file = self.outprefix + '.phandango.tre'

            if self.verbose:
                print('Making Phandango distance matrix', dist_matrix_file, flush=True)
            Summary._write_distance_matrix(matrix, dist_matrix_file)

            if self.verbose:
                print('Making Phandango tree file', tree_file, flush=True)
            Summary._newick_from_dist_matrix(dist_matrix_file, tree_file)
            os.unlink(dist_matrix_file)
        else:
            print('Made csv file. Not making Phandango files because only one sample remains after filtering', file=sys.stderr)

        if self.verbose:
            print('Finished', flush=True)



import tempfile
import shutil
import os
import pyfastaq
from ariba import common

class Error (Exception): pass

class Runner:
    def __init__(
      self,
      infile,
      outfile,
      seq_identity_threshold=0.9,
      threads=1,
      length_diff_cutoff=0.9,
      verbose=False,
      cd_hit_est='cd-hit-est',
      rename_suffix='x',
    ):

        if not os.path.exists(infile):
            raise Error('File not found: "' + infile + '". Cannot continue')

        self.infile = os.path.abspath(infile)
        self.outfile = os.path.abspath(outfile)
        self.seq_identity_threshold = seq_identity_threshold
        self.threads = threads
        self.length_diff_cutoff = length_diff_cutoff
        self.verbose = verbose
        self.cd_hit_est = cd_hit_est
        self.rename_suffix = rename_suffix


    def fake_run(self):
        '''Doesn't actually run cd-hit. Instead, puts each input sequence into its own cluster. So it's as if cdhit was run, but didn't cluster anything'''
        tmpdir = tempfile.mkdtemp(prefix='tmp.run_cd-hit.', dir=os.getcwd())
        tmp_fa = os.path.join(tmpdir, 'cdhit.fa')
        clusters = {}
        seq_reader = pyfastaq.sequences.file_reader(self.infile)
        f = pyfastaq.utils.open_file_write(tmp_fa)

        for seq in seq_reader:
            if seq.id in clusters:
                pyfastaq.utils.close(f)
                shutil.rmtree(tmpdir)
                raise Error('Sequence name "' + seq.id + '" not unique. Cannot continue')

            clusters[seq.id] = {seq.id}
            print(seq, file=f)

        pyfastaq.utils.close(f)
        clusters = self._rename_clusters(clusters, tmp_fa, self.outfile, rename_suffix=self.rename_suffix)
        shutil.rmtree(tmpdir)
        return clusters


    @staticmethod
    def _load_user_clusters_file(filename):
        f = pyfastaq.utils.open_file_read(filename)
        seq_to_cluster = {}
        for line in f:
            data = line.rstrip().split()

            for seq_name in data:
                if seq_name in seq_to_cluster:
                    pyfastaq.utils.close(f)
                    raise Error('Error reading clusters file. The sequence "' + seq_name + '" was found more than once in the file ' + filename)
                seq_to_cluster[seq_name] = data[0]

        pyfastaq.utils.close(f)
        return seq_to_cluster


    def run_get_clusters_from_file(self, infile):
        '''Instead of running cdhit, gets the clusters info from the input dict.
           Dict expected to be key=sequence name, value=name of cluster'''
        seq_to_cluster = self._load_user_clusters_file(infile)
        cluster_names = set(seq_to_cluster.values())
        tmpdir = tempfile.mkdtemp(prefix='tmp.run_cd-hit.', dir=os.getcwd())
        tmp_fa = os.path.join(tmpdir, 'cdhit.fa')
        clusters = {}
        seq_reader = pyfastaq.sequences.file_reader(self.infile)
        f = pyfastaq.utils.open_file_write(tmp_fa)

        for seq in seq_reader:
            if seq.id in clusters and seq.id in clusters[seq.id]:
                pyfastaq.utils.close(f)
                shutil.rmtree(tmpdir)
                raise Error('Sequence name "' + seq.id + '" not unique. Cannot continue')

            if seq.id not in seq_to_cluster:
                raise Error('Error forcing cdhit clustering. Found sequence ' + seq.id + ' in FASTA file, but not in provided clusters info from file ' + infile)

            cluster = seq_to_cluster[seq.id]
            if cluster not in clusters:
                clusters[cluster] = set()

            clusters[cluster].add(seq.id)
            if seq.id in cluster_names:
                print(seq, file=f)

        pyfastaq.utils.close(f)
        clusters = self._rename_clusters(clusters, tmp_fa, self.outfile, rename_suffix=self.rename_suffix)
        shutil.rmtree(tmpdir)
        return clusters


    def _get_ids(self, infile):
        seq_reader = pyfastaq.sequences.file_reader(infile)
        return set([seq.id for seq in seq_reader])


    @staticmethod
    def _parse_cluster_info_file(infile, cluster_representatives):
        f = pyfastaq.utils.open_file_read(infile)
        cluster_sets = {}
        found_representatives = {}  # store cluster number -> representative name

        for line in f:
            data = line.rstrip().split()
            seqname = data[2]
            if not (seqname.startswith('>') and seqname.endswith('...')):
                raise Error('Unexpected format of line from cdhit output file "' + infile + '". Line is:\n' + line)
            seqname = seqname[1:-3]

            cluster_number = int(data[0]) # this is the cluster number used by cdhit
            if cluster_number not in cluster_sets:
                cluster_sets[cluster_number] = set()

            cluster_sets[cluster_number].add(seqname)

            if data[3] == '*':
                found_representatives[cluster_number] = seqname

        pyfastaq.utils.close(f)

        if set(found_representatives.values()) != cluster_representatives:
            raise Error('Mismatch in cdhit output sequence names between fasta file and clusters file. Cannot continue')

        clusters = {}
        for cluster_number, cluster_name in found_representatives.items():
            clusters[cluster_name] = cluster_sets[cluster_number]

        return clusters


    @staticmethod
    def _rename_clusters(clusters_dict, infile, outfile, rename_suffix='x'):
        new_clusters_dict = {}
        freader = pyfastaq.sequences.file_reader(infile)
        f_out = pyfastaq.utils.open_file_write(outfile)

        for seq in freader:
            original_name = seq.id
            assert original_name in clusters_dict
            new_name = original_name.split('.')[0] + '.' + rename_suffix

            if new_name in new_clusters_dict:
                suffix = 2
                while new_name + '.' + str(suffix) in new_clusters_dict:
                    suffix += 1
                new_name += '.' + str(suffix)

            new_clusters_dict[new_name] = clusters_dict[original_name]
            seq.id = new_name
            print(seq, file=f_out)

        pyfastaq.utils.close(f_out)

        return new_clusters_dict


    def run(self):
        tmpdir = tempfile.mkdtemp(prefix='tmp.run_cd-hit.', dir=os.getcwd())
        cdhit_fasta = os.path.join(tmpdir, 'cdhit')
        cluster_info_outfile = cdhit_fasta + '.bak.clstr'

        cmd = ' '.join([
            self.cd_hit_est,
            '-i', self.infile,
            '-o', cdhit_fasta,
            '-c', str(self.seq_identity_threshold),
            '-T', str(self.threads),
            '-s', str(self.length_diff_cutoff),
            '-d 0',
            '-bak 1',
        ])

        common.syscall(cmd, verbose=self.verbose)
        cluster_representatives = self._get_ids(cdhit_fasta)
        clusters = self._parse_cluster_info_file(cluster_info_outfile, cluster_representatives)
        clusters = self._rename_clusters(clusters, cdhit_fasta, self.outfile, rename_suffix=self.rename_suffix)

        shutil.rmtree(tmpdir)
        return clusters




import argparse
from ariba import ref_genes_getter


def run():
    allowed_dbs = ['argannot', 'card', 'resfinder','vfdb']
    parser = argparse.ArgumentParser(
        description = 'Downloads reference data',
        usage = 'ariba getref [options] <' + '|'.join(allowed_dbs) + '> <outprefix>'
    )

    parser.add_argument('--genetic_code', type=int, help='Number of genetic code to use. Currently supported 1,4,11 [%(default)s]', choices=[1,4,11], default=11, metavar='INT')
    parser.add_argument('db', help='Database to download. Must be one of: ' + ' '.join(allowed_dbs), choices=allowed_dbs)
    parser.add_argument('outprefix', help='Prefix of output filenames')
    options = parser.parse_args()

    getter = ref_genes_getter.RefGenesGetter(options.db, genetic_code=options.genetic_code)
    getter.run(options.outprefix)




import sys
import argparse
from ariba import ref_preparer, external_progs, versions

def run():
    parser = argparse.ArgumentParser(
        description = 'ARIBA: Antibiotic Resistance Identification By Assembly',
        usage = 'ariba prepareref [options] <outdir>',
        epilog = 'REQUIRED: either --ref_prefix, or at least one of --presabs, --varonly, --noncoding')
    input_group = parser.add_argument_group('input files options')
    input_group.add_argument('--ref_prefix', help='Prefix of input files (same as was used with getref), to save listing --preseabs,--varonly ...etc. Will look for files called "ref_prefix." followed by: metadata.tsv,presence_absence.fa,noncoding.fa,variants_only.fa. Using this will cause these to be ignored if used: --presabs,--varonly,--noncoding,--metadata', metavar='FILENAME_PREFIX')
    input_group.add_argument('--presabs', help='FASTA file of presence absence genes', metavar='FILENAME')
    input_group.add_argument('--varonly', help='FASTA file of variants only genes', metavar='FILENAME')
    input_group.add_argument('--noncoding', help='FASTA file of noncoding sequences', metavar='FILENAME')
    input_group.add_argument('--metadata', help='tsv file of metadata about the reference sequences', metavar='FILENAME')

    cdhit_group = parser.add_argument_group('cd-hit options')
    cdhit_group.add_argument('--no_cdhit', action='store_true', help='Do not run cd-hit. Each input sequence is put into its own "cluster". Incompatible with --cdhit_clusters.')
    cdhit_group.add_argument('--cdhit_clusters', help='File specifying how the sequences should be clustered. Will be used instead of running cdhit. Format is one cluster per line. Sequence names separated by whitespace. First name in line is the cluster representative. Incompatible with --no_cdhit', metavar='FILENAME')
    cdhit_group.add_argument('--cdhit_min_id', type=float, help='Sequence identity threshold (cd-hit option -c) [%(default)s]', default=0.9, metavar='FLOAT')
    cdhit_group.add_argument('--cdhit_min_length', type=float, help='length difference cutoff (cd-hit option -s) [%(default)s]', default=0.9, metavar='FLOAT')

    other_group = parser.add_argument_group('other options')
    other_group.add_argument('--min_gene_length', type=int, help='Minimum allowed length in nucleotides of reference genes [%(default)s]', metavar='INT', default=6)
    other_group.add_argument('--max_gene_length', type=int, help='Maximum allowed length in nucleotides of reference genes [%(default)s]', metavar='INT', default=10000)
    other_group.add_argument('--genetic_code', type=int, help='Number of genetic code to use. Currently supported 1,4,11 [%(default)s]', choices=[1,4,11], default=11, metavar='INT')
    other_group.add_argument('--threads', type=int, help='Number of threads (currently only applies to cdhit) [%(default)s]', default=1, metavar='INT')
    other_group.add_argument('--verbose', action='store_true', help='Be verbose')

    parser.add_argument('outdir', help='Output directory (must not already exist)')
    options = parser.parse_args()

    if options.no_cdhit and options.cdhit_clusters is not None:
        sys.exit('Cannot use both --no_cdhit and --cdhit_clusters. Neither or exactly one of those options must be used')

    extern_progs, version_report_lines = versions.get_all_versions()
    if options.verbose:
        print(*version_report_lines, sep='\n')

    preparer = ref_preparer.RefPreparer(
        extern_progs,
        version_report_lines=version_report_lines,
        ref_prefix=options.ref_prefix,
        presabs=options.presabs,
        varonly=options.varonly,
        noncoding=options.noncoding,
        metadata=options.metadata,
        min_gene_length=options.min_gene_length,
        max_gene_length=options.max_gene_length,
        genetic_code=options.genetic_code,
        cdhit_min_id=options.cdhit_min_id,
        cdhit_min_length=options.cdhit_min_length,
        run_cdhit=not options.no_cdhit,
        clusters_file=options.cdhit_clusters,
        threads=options.threads,
        verbose=options.verbose,
    )

    preparer.run(options.outdir)



import sys
import os
import pickle
from ariba import common, mapping, reference_data

class Error (Exception): pass


class RefPreparer:
    def __init__(self,
        extern_progs,
        version_report_lines=None,
        ref_prefix=None,
        presabs=None,
        varonly=None,
        noncoding=None,
        metadata=None,
        min_gene_length=6,
        max_gene_length=10000,
        genetic_code=11,
        cdhit_min_id=0.9,
        cdhit_min_length=0.9,
        run_cdhit=True,
        clusters_file=None,
        threads=1,
        verbose=False,
    ):
        self.extern_progs = extern_progs

        if version_report_lines is None:
            self.version_report_lines = []
        else:
            self.version_report_lines = version_report_lines

        self.ref_prefix = ref_prefix
        self.presabs = presabs
        self.varonly = varonly
        self.noncoding = noncoding
        self.metadata = metadata
        self.min_gene_length = min_gene_length
        self.max_gene_length = max_gene_length
        self.genetic_code = genetic_code
        self.cdhit_min_id = cdhit_min_id
        self.cdhit_min_length = cdhit_min_length
        self.run_cdhit = run_cdhit
        self.clusters_file = clusters_file
        self.threads = threads
        self.verbose = verbose


    @staticmethod
    def _get_ref_files(ref_prefix, presabs, varonly, noncoding, metadata, verbose=False):
        if {None} == {ref_prefix, presabs, varonly, noncoding}:
            raise Error('Error in RefPreparer._get_ref_files. All input files and ref_prefix were None. Cannot continue')

        filenames = {
            'presabs': presabs,
            'varonly': varonly,
            'noncoding': noncoding,
            'metadata': metadata,
        }

        file_suffixes = {
            'presabs': 'presence_absence.fa',
            'varonly': 'variants_only.fa',
            'noncoding': 'noncoding.fa',
            'metadata': 'metadata.tsv',
        }

        if verbose:
            print('\nLooking for input files ...')

        for key in file_suffixes:
            if ref_prefix is not None:
                filename = os.path.abspath(ref_prefix + '.' + file_suffixes[key])

                if os.path.exists(filename):
                    if verbose:
                        print('Found: ', filename, '.\n    ...treating it as if this was used: --', key, ' ', filename, sep='')
                    filenames[key] = filename
                else:
                    if verbose:
                        print('Not found:', filename)
                    filenames[key] = None
            elif filenames[key] is not None:
                if os.path.exists(filenames[key]):
                    filenames[key] = os.path.abspath(filenames[key])
                    if verbose:
                        print('Found: ', filenames[key], ' from option --', key, sep='')
                else:
                    raise Error('File not found! Cannot continue. Looked for: ' + filenames[key])

        if {None} == {filenames['presabs'], filenames['varonly'], filenames['noncoding']}:
            raise Error('Error in RefPreparer._get_ref_files. No FASTA files given! Cannot continue')

        return filenames


    def _write_info_file(self, outfile):
        with open(outfile, 'w') as fout:
            for key in ('presabs', 'varonly', 'noncoding', 'metadata'):
                print('input_' + key, self.filenames[key], sep='\t', file=fout)

            print('genetic_code', self.genetic_code, sep='\t', file=fout)


    def run(self, outdir):
        original_dir = os.getcwd()

        if os.path.exists(outdir):
            raise Error('Error! Output directory ' + outdir + ' already exists. Cannot continue')

        try:
            os.mkdir(outdir)
        except:
            raise Error('Error making output directory ' + outdir + '. Cannot continue')

        with open(os.path.join(outdir, 'version_info.txt'), 'w') as f:
            print('ARIBA run with this command:', file=f)
            print(' '.join([sys.argv[0]] + ['prepareref'] + sys.argv[1:]), file=f)
            print('from this directory:', original_dir, file=f)
            print(file=f)
            print(*self.version_report_lines, sep='\n', file=f)

        self.filenames = self._get_ref_files(self.ref_prefix, self.presabs, self.varonly, self.noncoding, self.metadata, self.verbose)
        self._write_info_file(os.path.join(outdir, 'info.txt'))

        self.refdata = reference_data.ReferenceData(
            presence_absence_fa=self.filenames['presabs'],
            variants_only_fa=self.filenames['varonly'],
            non_coding_fa=self.filenames['noncoding'],
            metadata_tsv=self.filenames['metadata'],
            min_gene_length=self.min_gene_length,
            max_gene_length=self.max_gene_length,
            genetic_code=self.genetic_code,
        )

        if self.verbose:
            print('\nLoading and checking input data', flush=True)

        refdata_outprefix = os.path.join(outdir, 'refcheck')
        self.refdata.rename_sequences(refdata_outprefix + '.rename_info')
        self.refdata.sanity_check(refdata_outprefix)

        if self.verbose:
            print('\nRunning cdhit', flush=True)
        cdhit_outprefix = os.path.join(outdir, 'cdhit')

        clusters = self.refdata.cluster_with_cdhit(
            refdata_outprefix + '.01.check_variants',
            cdhit_outprefix,
            seq_identity_threshold=self.cdhit_min_id,
            threads=self.threads,
            length_diff_cutoff=self.cdhit_min_length,
            nocluster=not self.run_cdhit,
            verbose=self.verbose,
            cd_hit_est=self.extern_progs.exe('cdhit'),
            clusters_file=self.clusters_file,
        )

        if self.verbose:
            print('\nWriting clusters to file.', len(clusters), 'in total', flush=True)

        clusters_pickle_file = cdhit_outprefix + '.clusters.pickle'
        with open(clusters_pickle_file, 'wb') as f:
            pickle.dump(clusters, f)

        cluster_representatives_fa = cdhit_outprefix + '.cluster_representatives.fa'

        if self.verbose:
            print('\nRunning bowtie2-build on FASTA of cluster representatives')

        mapping.bowtie2_index(
            cluster_representatives_fa,
            cluster_representatives_fa,
            bowtie2=self.extern_progs.exe('bowtie2'),
            verbose=self.verbose,
        )

        if self.verbose:
            print('\nRunning samtools faidx on FASTA of cluster representatives')

        cmd = ' '.join([
            self.extern_progs.exe('samtools'),
            'faidx',
            cluster_representatives_fa
        ])

        common.syscall(cmd, verbose=self.verbose)
